
#pragma once

struct lambda_item
{
   lambda_item()
      : ref_counter_(0)
   {}

   virtual ~lambda_item() = 0
   {}

   virtual void print(std::wostream & wos) = 0;
   virtual bool is_free_variable(std::wstring const & name) = 0;

private:
   lambda_item(lambda_item const &);
   lambda_item & operator=(lambda_item const &);

private:
   unsigned ref_counter_;

   friend void intrusive_ptr_add_ref(lambda_item * p);
   friend void intrusive_ptr_release(lambda_item * p);
};

void intrusive_ptr_add_ref(lambda_item * p)
{
   ++p->ref_counter_;
}

void intrusive_ptr_release(lambda_item * p)
{
   if (--p->ref_counter_ == 0)
      delete p;
}

typedef boost::intrusive_ptr<lambda_item> lambda_item_ptr;

struct variable : lambda_item
{
   variable(std::wstring const & name)
      : name_(name)
   {}

   void print(std::wostream & wos)
   {
      wos << name_;
   }

   bool is_free_variable(std::wstring const & name)
   {
      return name_ == name;
   }

   std::wstring name_;
};

typedef boost::intrusive_ptr<variable> variable_ptr;

struct application : lambda_item
{
   application(lambda_item_ptr function, lambda_item_ptr arg)
      : function_(function)
      , arg_(arg)
   {}

   void print(std::wostream & wos)
   {
      wos << L"(";
      function_->print(wos);
      wos << L" ";
      arg_->print(wos);
      wos << L")";
   }

   bool is_free_variable(std::wstring const & name)
   {
      return function_->is_free_variable(name) ||
                  arg_->is_free_variable(name);
   }

   lambda_item_ptr function_;
   lambda_item_ptr arg_;
};

struct abstraction : lambda_item
{
   abstraction(variable_ptr var, lambda_item_ptr body)
      : var_(var)
      , body_(body)
   {}

   void print(std::wostream & wos)
   {
      wos << L"(\\";
      var_->print(wos);
      wos << L" ";
      body_->print(wos);
      wos << L")";
   }

   bool is_free_variable(std::wstring const & name)
   {
      if (var_->name_ == name)
         return false;

      return body_->is_free_variable(name);
   }

   variable_ptr var_;
   lambda_item_ptr body_;
};
